@@ -91,7 +91,7 @@ static int save_certificate_home(X509 * cert);
  * Global variables
  */
 static char *root_url = NULL;  /*Holds the URL we are connecting to*/
-static SockHandler *sh;
+static Dsh *sh;
 
 
 #ifdef ENABLE_SSL
@@ -107,8 +107,8 @@ static int dialog_get_answer_number(void)
    char *dpip_tag, *response;
 
    /* Read the dpi command from STDIN */
-   dpip_tag = sock_handler_read(sh);
-   response = a_Dpip_get_attr(dpip_tag, strlen(dpip_tag), "msg");
+   dpip_tag = a_Dpip_dsh_read_token(sh, 1);
+   response = a_Dpip_get_attr(dpip_tag, "msg");
    response_number = (response) ? strtol (response, NULL, 10) : -1;
    dFree(dpip_tag);
    dFree(response);
@@ -130,9 +130,9 @@ static void yes_ssl_support(void)
    SSL * ssl_connection = NULL;
 
    char *dpip_tag = NULL, *cmd = NULL, *url = NULL, *http_query = NULL,
-        *proxy_url = NULL, *proxy_connect = NULL;
+        *proxy_url = NULL, *proxy_connect = NULL, *check_cert = NULL;
    char buf[4096];
-   int retval = 0;
+   int ret = 0;
    int network_socket = -1;
 
 
@@ -193,15 +193,19 @@ static void yes_ssl_support(void)
       SSL_set_verify(ssl_connection, SSL_VERIFY_NONE, 0);
 
       /*Get the network address and command to be used*/
-      dpip_tag = sock_handler_read(sh);
-      cmd = a_Dpip_get_attr(dpip_tag, strlen(dpip_tag), "cmd");
-      proxy_url = a_Dpip_get_attr(dpip_tag, strlen(dpip_tag), "proxy_url");
+      dpip_tag = a_Dpip_dsh_read_token(sh, 1);
+      cmd = a_Dpip_get_attr(dpip_tag, "cmd");
+      proxy_url = a_Dpip_get_attr(dpip_tag, "proxy_url");
       proxy_connect =
-                  a_Dpip_get_attr(dpip_tag, strlen(dpip_tag), "proxy_connect");
-      url = a_Dpip_get_attr(dpip_tag, strlen(dpip_tag), "url");
-      http_query = a_Dpip_get_attr(dpip_tag, strlen(dpip_tag), "query");
+                  a_Dpip_get_attr(dpip_tag, "proxy_connect");
+      url = a_Dpip_get_attr(dpip_tag, "url");
+      http_query = a_Dpip_get_attr(dpip_tag, "query");
+      if (!(check_cert = a_Dpip_get_attr(dpip_tag, "check_cert"))) {
+         /* allow older dillo versions use this dpi */
+         check_cert = dStrdup("true");
+      }
 
-      if (cmd == NULL || url == NULL || http_query == NULL){
+      if (!cmd || !url || !http_query) {
          MSG("***Value of cmd, url or http_query is NULL"
                     " - cannot continue\n");
          exit_error = 1;
@@ -288,7 +292,8 @@ static void yes_ssl_support(void)
 
    /*Use handle error function to decide what to do*/
    if (exit_error == 0){
-      if (handle_certificate_problem(ssl_connection) < 0){
+      if (strcmp(check_cert, "true") == 0 &&
+          handle_certificate_problem(ssl_connection) < 0){
          MSG("Certificate verification error\n");
          exit_error = 1;
       }
@@ -304,14 +309,14 @@ static void yes_ssl_support(void)
 
       /*Send dpi command*/
       d_cmd = a_Dpip_build_cmd("cmd=%s url=%s", "start_send_page", url);
-      sock_handler_write_str(sh, 1, d_cmd);
+      a_Dpip_dsh_write_str(sh, 1, d_cmd);
       dFree(d_cmd);
 
       /*Send remaining data*/
 
-      while ((retval = SSL_read(ssl_connection, buf, 4096)) > 0 ){
+      while ((ret = SSL_read(ssl_connection, buf, 4096)) > 0 ){
          /* flush is good for dialup speed */
-         sock_handler_write(sh, 1, buf, (size_t)retval);
+         a_Dpip_dsh_write(sh, 1, buf, (size_t)ret);
       }
    }
 
@@ -322,6 +327,7 @@ static void yes_ssl_support(void)
    dFree(http_query);
    dFree(proxy_url);
    dFree(proxy_connect);
+   dFree(check_cert);
 
    if (network_socket != -1){
       close(network_socket);
@@ -379,7 +385,7 @@ static int get_network_connection(char * url)
    root_url = dStrdup(url_look_up);
    hp=gethostbyname(url_look_up);
 
-   /*url_look_uip no longer needed, so free if neccessary*/
+   /*url_look_uip no longer needed, so free if necessary*/
    if (url_look_up_length != 0){
       dFree(url_look_up);
    }
@@ -409,14 +415,17 @@ static int get_network_connection(char * url)
  * allow the user to decide what to do.  It may save the
  * certificate to the user's .dillo directory if it is
  * trusted.
+ *
+ * TODO: Rearrange this to get rid of redundancy.
+ *
  * Return value: -1 on abort, 0 or higher on continue
  */
 static int handle_certificate_problem(SSL * ssl_connection)
 {
    int response_number;
-   int retval = -1;
+   int ret = -1;
    long st;
-   char *cn, *cn_end;
+   char *cn;
    char buf[4096], *d_cmd, *msg;
 
    X509 * remote_cert;
@@ -431,7 +440,7 @@ static int handle_certificate_problem(SSL * ssl_connection)
          "This site CAN NOT be trusted. Sending data is NOT SAFE.\n"
          "What do I do?",
          "Continue", "Cancel");
-      sock_handler_write_str(sh, 1, d_cmd);
+      a_Dpip_dsh_write_str(sh, 1, d_cmd);
       dFree(d_cmd);
 
       /*Read the user's response*/
@@ -439,7 +448,7 @@ static int handle_certificate_problem(SSL * ssl_connection)
 
       /*Abort on anything but "Continue"*/
       if (response_number == 1){
-         retval = 0;
+         ret = 0;
       }
 
    } else {
@@ -447,36 +456,37 @@ static int handle_certificate_problem(SSL * ssl_connection)
       st = SSL_get_verify_result(ssl_connection);
       switch (st) {
       case X509_V_OK:      /*Everything is Kosher*/
-         retval = 0;
+         ret = 0;
          break;
       case X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT:
          /*Either self signed and untrusted*/
          /*Extract CN from certificate name information*/
-         cn = strstr(remote_cert->name, "/CN=") + 4;
-         if (cn == NULL)
-            break;
+         if ((cn = strstr(remote_cert->name, "/CN=")) == NULL) {
+            strcpy(buf, "(no CN given)");
+         } else {
+            char *cn_end;
 
-         if ((cn_end = strstr(cn, "/")) == NULL )
-            cn_end = cn + strlen(cn);
+            cn += 4;
 
-         strncpy(buf, cn, (size_t) (cn_end - cn));
-
-         /*Add terminating NULL*/
-         buf[cn_end - cn] = 0;
+            if ((cn_end = strstr(cn, "/")) == NULL )
+               cn_end = cn + strlen(cn);
 
+            strncpy(buf, cn, (size_t) (cn_end - cn));
+            buf[cn_end - cn] = '\0';
+         }
          msg = dStrconcat("The remote certificate is self-signed and "
                           "untrusted.\nFor address: ", buf, NULL);
          d_cmd = a_Dpip_build_cmd(
             "cmd=%s msg=%s alt1=%s alt2=%s alt3=%s",
-            "dialog", msg, "Continue", "Cancel", "Trust Certificate");
-         sock_handler_write_str(sh, 1, d_cmd);
+            "dialog", msg, "Continue", "Cancel", "Save Certificate");
+         a_Dpip_dsh_write_str(sh, 1, d_cmd);
          dFree(d_cmd);
          dFree(msg);
 
          response_number = dialog_get_answer_number();
          switch (response_number){
             case 1:
-               retval = 0;
+               ret = 0;
                break;
             case 2:
                break;
@@ -485,7 +495,7 @@ static int handle_certificate_problem(SSL * ssl_connection)
                /*Potential security problems because we are writing
                 *to the filesystem*/
                save_certificate_home(remote_cert);
-               retval = 1;
+               ret = 1;
                break;
             default:
                break;
@@ -499,12 +509,12 @@ static int handle_certificate_problem(SSL * ssl_connection)
             "The issuer for the remote certificate cannot be found\n"
             "The authenticity of the remote certificate cannot be trusted",
             "Continue", "Cancel");
-         sock_handler_write_str(sh, 1, d_cmd);
+         a_Dpip_dsh_write_str(sh, 1, d_cmd);
          dFree(d_cmd);
 
          response_number = dialog_get_answer_number();
          if (response_number == 1) {
-            retval = 0;
+            ret = 0;
          }
          break;
 
@@ -518,12 +528,12 @@ static int handle_certificate_problem(SSL * ssl_connection)
             "The remote certificate signature could not be read\n"
             "or is invalid and should not be trusted",
             "Continue", "Cancel");
-         sock_handler_write_str(sh, 1, d_cmd);
+         a_Dpip_dsh_write_str(sh, 1, d_cmd);
          dFree(d_cmd);
 
          response_number = dialog_get_answer_number();
          if (response_number == 1) {
-            retval = 0;
+            ret = 0;
          }
          break;
       case X509_V_ERR_CERT_NOT_YET_VALID:
@@ -537,12 +547,12 @@ static int handle_certificate_problem(SSL * ssl_connection)
             "presented has a starting validity after today's date\n"
             "You should be cautious about using this site",
             "Continue", "Cancel");
-         sock_handler_write_str(sh, 1, d_cmd);
+         a_Dpip_dsh_write_str(sh, 1, d_cmd);
          dFree(d_cmd);
 
          response_number = dialog_get_answer_number();
          if (response_number == 1) {
-            retval = 0;
+            ret = 0;
          }
          break;
       case X509_V_ERR_CERT_HAS_EXPIRED:
@@ -554,11 +564,11 @@ static int handle_certificate_problem(SSL * ssl_connection)
             "wasn't designed to last this long. You should avoid \n"
             "this site.",
             "Continue", "Cancel");
-         sock_handler_write_str(sh, 1, d_cmd);
+         a_Dpip_dsh_write_str(sh, 1, d_cmd);
          dFree(d_cmd);
          response_number = dialog_get_answer_number();
          if (response_number == 1) {
-            retval = 0;
+            ret = 0;
          }
          break;
       case X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD:
@@ -573,11 +583,11 @@ static int handle_certificate_problem(SSL * ssl_connection)
             "making it impossible to determine if the certificate\n"
             "is valid.  You should not trust this certificate.",
             "Continue", "Cancel");
-         sock_handler_write_str(sh, 1, d_cmd);
+         a_Dpip_dsh_write_str(sh, 1, d_cmd);
          dFree(d_cmd);
          response_number = dialog_get_answer_number();
          if (response_number == 1) {
-            retval = 0;
+            ret = 0;
          }
          break;
       case X509_V_ERR_INVALID_CA:
@@ -593,11 +603,11 @@ static int handle_certificate_problem(SSL * ssl_connection)
             "with the remote system.  The connection should not\n"
             "be trusted",
             "Continue", "Cancel");
-         sock_handler_write_str(sh, 1, d_cmd);
+         a_Dpip_dsh_write_str(sh, 1, d_cmd);
          dFree(d_cmd);
          response_number = dialog_get_answer_number();
          if (response_number == 1) {
-            retval = 0;
+            ret = 0;
          }
          break;
       case X509_V_ERR_SUBJECT_ISSUER_MISMATCH:
@@ -608,29 +618,63 @@ static int handle_certificate_problem(SSL * ssl_connection)
             "dialog",
             "Some of the information presented by the remote system\n"
             "does not match other information presented\n"
-            "This may be an attempt to evesdrop on communications",
+            "This may be an attempt to eavesdrop on communications",
             "Continue", "Cancel");
-         sock_handler_write_str(sh, 1, d_cmd);
+         a_Dpip_dsh_write_str(sh, 1, d_cmd);
          dFree(d_cmd);
+         response_number = dialog_get_answer_number();
+         if (response_number == 1) {
+            ret = 0;
+         }
+         break;
+      case X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN:
+         d_cmd = a_Dpip_build_cmd(
+            "cmd=%s msg=%s alt1=%s alt2=%s",
+            "dialog",
+            "Self signed certificate in certificate chain. The certificate "
+            "chain could be built up using the untrusted certificates but the "
+            "root could not be found locally.",
+            "Continue", "Cancel");
+         a_Dpip_dsh_write_str(sh, 1, d_cmd);
+         dFree(d_cmd);
+         response_number = dialog_get_answer_number();
+         if (response_number == 1) {
+            ret = 0;
+         }
+         break;
+      case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:
+         d_cmd = a_Dpip_build_cmd(
+            "cmd=%s msg=%s alt1=%s alt2=%s",
+            "dialog",
+            "Unable to get local issuer certificate. The issuer certificate "
+            "of an untrusted certificate cannot be found.",
+            "Continue", "Cancel");
+         a_Dpip_dsh_write_str(sh, 1, d_cmd);
+         dFree(d_cmd);
+         response_number = dialog_get_answer_number();
+         if (response_number == 1) {
+            ret = 0;
+         }
+         break;
       default:             /*Need to add more options later*/
          snprintf(buf, 80,
                   "The remote certificate cannot be verified (code %ld)", st);
          d_cmd = a_Dpip_build_cmd(
             "cmd=%s msg=%s alt1=%s alt2=%s",
             "dialog", buf, "Continue", "Cancel");
-         sock_handler_write_str(sh, 1, d_cmd);
+         a_Dpip_dsh_write_str(sh, 1, d_cmd);
          dFree(d_cmd);
          response_number = dialog_get_answer_number();
          /*abort on anything but "Continue"*/
          if (response_number == 1){
-            retval = 0;
+            ret = 0;
          }
       }
       X509_free(remote_cert);
       remote_cert = 0;
    }
 
-   return retval;
+   return ret;
 }
 
 /*
@@ -643,7 +687,7 @@ static int save_certificate_home(X509 * cert)
 
    FILE * fp = NULL;
    uint_t i = 0;
-   int retval = 1;
+   int ret = 1;
 
    /*Attempt to create .dillo/certs blindly - check later*/
    snprintf(buf,4096,"%s/.dillo/", dGethomedir());
@@ -666,7 +710,7 @@ static int save_certificate_home(X509 * cert)
             PEM_write_X509(fp, cert);
             fclose(fp);
             MSG("Wrote certificate\n");
-            retval = 0;
+            ret = 0;
             break;
          }
       } else {
@@ -676,7 +720,7 @@ static int save_certificate_home(X509 * cert)
       /*Don't loop too many times - just give up*/
    } while (i < 1024);
 
-   return retval;
+   return ret;
 }
 
 
@@ -694,14 +738,14 @@ static void no_ssl_support(void)
    char *d_cmd;
 
    /* Read the dpi command from STDIN */
-   dpip_tag = sock_handler_read(sh);
+   dpip_tag = a_Dpip_dsh_read_token(sh, 1);
 
    MSG("{In https.filter.dpi}\n");
    MSG("no_ssl_support version\n");
 
-   cmd = a_Dpip_get_attr(dpip_tag, strlen(dpip_tag), "cmd");
-   url = a_Dpip_get_attr(dpip_tag, strlen(dpip_tag), "url");
-   http_query = a_Dpip_get_attr(dpip_tag, strlen(dpip_tag), "query");
+   cmd = a_Dpip_get_attr(dpip_tag, "cmd");
+   url = a_Dpip_get_attr(dpip_tag, "url");
+   http_query = a_Dpip_get_attr(dpip_tag, "query");
 
    MSG("{ cmd: %s}\n", cmd);
    MSG("{ url: %s}\n", url);
@@ -710,14 +754,14 @@ static void no_ssl_support(void)
    MSG("{ sending dpip cmd...}\n");
 
    d_cmd = a_Dpip_build_cmd("cmd=%s url=%s", "start_send_page", url);
-   sock_handler_write_str(sh, 1, d_cmd);
+   a_Dpip_dsh_write_str(sh, 1, d_cmd);
    dFree(d_cmd);
 
    MSG("{ dpip cmd sent.}\n");
 
    MSG("{ sending HTML...}\n");
 
-   sock_handler_printf(sh, 1,
+   a_Dpip_dsh_printf(sh, 1,
       "Content-type: text/html\n\n"
       "<!DOCTYPE HTML PUBLIC '-//W3C//DTD HTML 4.01 Transitional//EN'>\n"
       "<html><head><title>SSL support is disabled</title></head>\n"
@@ -758,8 +802,19 @@ static void no_ssl_support(void)
 /*---------------------------------------------------------------------------*/
 int main(void)
 {
+   char *dpip_tag;
+
    /* Initialize the SockHandler for this filter dpi */
-   sh = sock_handler_new(STDIN_FILENO, STDOUT_FILENO, 8*1024);
+   sh = a_Dpip_dsh_new(STDIN_FILENO, STDOUT_FILENO, 8*1024);
+
+   /* Authenticate our client... */
+   if (!(dpip_tag = a_Dpip_dsh_read_token(sh, 1)) ||
+       a_Dpip_check_auth(dpip_tag) < 0) {
+      MSG("can't authenticate request: %s\n", dStrerror(errno));
+      a_Dpip_dsh_close(sh);
+      return 1;
+   }
+   dFree(dpip_tag);
 
 #ifdef ENABLE_SSL
    yes_ssl_support();
@@ -768,8 +823,8 @@ int main(void)
 #endif
 
    /* Finish the SockHandler */
-   sock_handler_close(sh);
-   sock_handler_free(sh);
+   a_Dpip_dsh_close(sh);
+   a_Dpip_dsh_free(sh);
 
    dFree(root_url);
 
