@@ -61,7 +61,7 @@ const char *const AboutSplash=
 "    <table border='0' cellspacing='0' cellpadding='2'><tr>\n"
 "    <td>\n"
 "    <td>\n"
-"     <a href='http://www.dillo.org/dillo2-help.html'>\n"
+"     <a href='http://www.dillo.org/dillo3-help.html'>\n"
 "     Help</a>\n"
 "    <tr>\n"
 "    <td>&nbsp;&nbsp;\n"
@@ -125,11 +125,16 @@ const char *const AboutSplash=
 "    <tr>\n"
 "    <td>&nbsp;&nbsp;\n"
 "    <td>\n"
+"     <a href='http://www.mathaba.net/'>Mathaba</a>\n"
+"    <tr>\n"
+"    <tr>\n"
+"    <td>&nbsp;&nbsp;\n"
+"    <td>\n"
 "     <a href='http://www.commondreams.org/'>C.&nbsp;Dreams</a>\n"
 "    <tr>\n"
 "    <td>&nbsp;&nbsp;\n"
 "    <td>\n"
-"     <a href='http://www.voltairenet.org/'>VoltaireNet</a>\n"
+"     <a href='http://www.voltairenet.org/en'>VoltaireNet</a>\n"
 "    <tr>\n"
 "    <td>&nbsp;&nbsp;\n"
 "    <td>\n"
@@ -153,12 +158,6 @@ const char *const AboutSplash=
 "    <table border='0' cellspacing='0' cellpadding='5'><tr><td>\n"
 "    <table border='0' cellpadding='2'><tr>\n"
 "    <td>&nbsp;&nbsp;\n"
-"    <td><a href='http://www.google.com/'>Google</a>\n"
-"    <tr>\n"
-"    <td>&nbsp;&nbsp;\n"
-"    <td><a href='http://www.wikipedia.org/'>Wikipedia</a>\n"
-"    <tr>\n"
-"    <td>&nbsp;&nbsp;\n"
 "    <td><a href='http://www.gutenberg.org/'>P.&nbsp;Gutenberg</a>\n"
 "    <tr>\n"
 "    <td>&nbsp;&nbsp;\n"
@@ -240,20 +239,27 @@ const char *const AboutSplash=
 "<tr>\n"
 " <td bgcolor='#CCCCCC'>\n"
 "  <h4>Release overview</h4>\n"
-"  July 3, 2009\n"
+"  September 06, 2011\n"
 "<tr>\n"
 " <td bgcolor='#FFFFFF'>\n"
 "  <table border='0' cellspacing='0' cellpadding='5'>\n"
 "  <tr>\n"
 "   <td>\n"
 "<p>\n"
-"This release of Dillo improves image size checking. It builds upon version\n"
-"2.1, which brought significant improvements such as basic CSS support and\n"
-"configurable keybindings.\n"
+"Dillo-3.0 is a port to FLTK-1.3, which is big news because FLTK-1.3.0 was\n"
+"<a href='http://fltk.org/articles.php?L1086'>released</a>\n"
+"in June, clearing the way for Dillo to return to those distributions\n"
+"which had excluded Dillo2 due to FLTK2 never being officially released.\n"
+"<p>\n"
+"Dillo-3.0 also has plenty of improvements and bugfixes.\n"
+"<p>\n"
+"After this release, the core team will focus on implementing the CSS\n"
+"feature of floating elements. This will <em>greatly</em> improve dillo's\n"
+"web page rendering since many sites have adopted floats instead of tables.\n"
+"<p>\n"
+"The new dillo3 has shown excellent stability in our experience.\n"
+"The core team welcomes developers willing to join our workforce.\n"
 "<p>\n"
-"Remember that the dillo project uses a release model where every new\n"
-"version shall be better than the last.\n"
-"<EM>Keep up with the latest one!</EM>\n"
 "  </table>\n"
 "</table>\n"
 "</table>\n"
@@ -273,65 +279,40 @@ const char *const AboutSplash=
 "  <table border='0' cellspacing='0' cellpadding='5'>\n"
 "  <tr>\n"
 "   <td>\n"
-"2.1.1:<br>\n"
 "<ul>\n"
-"<li>Add additional size checks for images.\n"
-"<li>Added support for css colors of the form rgb(255, 255, 255).\n"
-"<li>Added the 'nop' keybinding (nop = NO_OPERATION; cancels a default hook).\n"
-"<li>Added 'stop' key action (not bound by default).\n"
-"<li>Reduced 'warning: ignoring return value of ...'\n"
-"</ul>\n"
-"2.1:<br>\n"
-"<ul>\n"
-"<li>Implemented basic CSS infrastructure!\n"
-"<li>Read user style from ~/.dillo/style.css.\n"
-"<li>Added configurable keybindings! (in ~/.dillo/keysrc)\n"
-"<li>Implemented \"search previous\" in string searches.\n"
-"<li>Ported the command line interface from dillo1\n"
-"<li>Set middle click to submit in a new TAB. (Helps to keep form data!)\n"
-"<li>Implemented Basic authentication!\n"
-"<li>Implemented a close-tab button for the GUI.\n"
-"<li>Implemented a tools menu.\n"
-"<li>Added dillo(1) man page.\n"
-"<li>Added \"font_max_size\", \"font_min_size\" dillorc options.\n"
-"<li>Added instant client-side redirects (aka. zero-delay META refresh).\n"
-"<li>Proxy support for HTTPS.\n"
-"<li>Updated the URL resolver to comply with RFC-3986.\n"
-"<li>Fixed Bookmarks modify's HTML so it wraps nicely on handhelds.\n"
-"<li>Made cookierc parsing more robust.\n"
-"<li>Fix: recover page focus when clicking outside of a widget.\n"
-"<li>Added support for the Q element. BUG#343\n"
-"<li>Added a right-click menu to form controls (show hiddens, submit, reset)\n"
-"<li>Added the \"http_language\" dillorc option for setting HTTP's Accept-Language.\n"
-"<li>Replace image loading button and page menu option with a tools menu option.\n"
-"<li>Implemented the \"overline\" text-decoration.\n"
-"<li>Enhanced and cleaned up text decorations for SUB and SUP.\n"
-"<li>Added \"View Stylesheets\" to the page menu.\n"
-"<li>System config files have moved to sysconfdir/dillo/\n"
-"<li>Allowed compilation with older machines by removing a few C99isms.\n"
-"<li>Switched SSL-enabled to configure.in (./configure --enable-ssl).\n"
-"<li>Removed redundant caller NULL checks already in the API.\n"
-"<li>Added use of inttypes.h when stdint.h isn't found.\n"
-"<li>Made the parser recognize \"[^ ]/>\"-terminated XML elements.\n"
-"<li>Brought in Sebastian's CSS parser from dillo-0.8.0-css-3.\n"
-"<li>Support CSS @import directive.\n"
-"<li>Improved CSS selector matching performance using hash tables.\n"
-"<li>Added support for descendant and child selectors.\n"
-"<li>Support selector specificity.\n"
-"<li>Replace bg_color dillorc option.\n"
-"<li>Remove text_color, link_color, and force_my_colors dillorc options.\n"
-"<li>Replace visited_color dillorc option.\n"
-"<li>Allow negative values for specific CSS properties only.\n"
-"<li>Disable negative margins for now as dw/* does not support them yet.\n"
-"<li>Disable form widgets while stylesheets are loading.\n"
-"<li>Implement --xid command line option (used by claws mail client).\n"
-"<li>Added the \"middle_click_drags_page\" dillorc option.\n"
-"<li>Set the File menu label to hide when the File menu-button is shown.\n"
-"<li>Made a big cleanup of cache.c WRT charset decoding (fixes bugs).\n"
-"<li>Made an extensive cleanup/fixup of the whole image handling process.\n"
-"<li>Fixed handling of META's content-type with no MIME type (e.g. only charset).\n"
-"<li>Added support for a quoted URL in META refresh.\n"
-"<li>Updated the GPL copyright note in the source files.\n"
+"<li>Ported Dillo to FLTK-1.3.\n"
+"<li>Native build on OSX.\n"
+"<li>Default binding for close-all changed from Alt-q to Ctrl-q.\n"
+"<li>Default binding for close-tab changed from Ctrl-q to Ctrl-w.\n"
+"<li>Default binding for left-tab changed to Shift-Ctrl-Tab.\n"
+"<li>Rewrote the User Interface: much simpler design and event handling.\n"
+"<li>Added on-the-fly panel resize (tiny/small/medium and normal/small icons).\n"
+"<li>'hide-panels' key action now hides the findbar if present,\n"
+"     and toggles display of the control panels otherwise.\n"
+"<li>Allow multiple search engines to be set in dillorc, with a menu\n"
+"    in the web search dialog to select between them.\n"
+"<li>Added an optional label to dillorc's search_url.\n"
+"    Format: \"[&lt;label&gt; ]&lt;url&gt;\"\n"
+"<li>Add right_click_closes_tab preference (default is middle click).\n"
+"<li>Allow binding to non-ASCII keys and multimedia keys.\n"
+"<li>Avoid a certificate dialog storm on some HTTPS sites (BUG#868).\n"
+"<li>Enable line wrapping for &lt;textarea&gt;. (BUG#903)\n"
+"<li>Avoid double render after going Back or Forward\n"
+"    (it takes half the time now!)\n"
+"<li>Implemented a custom tabs handler (to allow fine control of it).\n"
+"<li>Rewrote dw's crossing-events dispatcher (avoids redundant events).\n"
+"<li>Fixed a years old bug: stamped tooltips when scrolling with keyboard.\n"
+"<li>Fixed a border case in URL resolver: empty path + {query|fragment}\n"
+"     (BUG#948)\n"
+"<li>Cancel the expected URL after offering a download (BUG#982)\n"
+"<li>Eliminated a pack of 22 compiler warnings (gcc-4.6.1 amd64)\n"
+"<li>Removed 'large' option of panel_size preference.\n"
+"<li>Removed --enable-ansi configure option.\n"
+"<li>Limit saved cookie size.\n"
+"<li>Wrap image alt text.\n"
+"<li>Added support for CSS adjacent sibling selectors.\n"
+"<li>Fix redraw loops and reenabled limit_text_width dillorc option.\n"
+"<li>Collapse parent's and first child's top margin.\n"
 "</ul>\n"
 "  </table>\n"
 "</table>\n"
@@ -364,8 +345,8 @@ const char *const AboutSplash=
 "    and the bug meter).\n"
 " <li> Dillo behaves very nicely when browsing local files, images, and HTML.\n"
 "   It's also very good for Internet searching.\n"
-" <li> This release is mainly intended <strong>for developers</strong>\n"
-"   and <em>advanced users</em>.\n"
+" <li> This release is mainly intended for <strong>developers</strong>\n"
+"   and <strong>advanced users</strong>.\n"
 " <li> Frames, Java and Javascript are not supported.\n"
 "</ul>\n"
 "<br>\n"
