@@ -21,6 +21,7 @@
 #include <errno.h>              /* for errno */
 #include <stdlib.h>
 #include <fcntl.h>
+#include <assert.h>
 #include <sys/socket.h>         /* for lots of socket stuff */
 #include <netinet/in.h>         /* for ntohl and stuff */
 #include <arpa/inet.h>          /* for inet_ntop */
@@ -47,27 +48,58 @@ D_STMT_START {                                                        \
 
 #define _MSG_BW(web, root, ...)
 
+static const int HTTP_SOCKET_USE_PROXY   = 0x1;
+static const int HTTP_SOCKET_QUEUED      = 0x4;
+static const int HTTP_SOCKET_TO_BE_FREED = 0x8;
+
 /* 'Url' and 'web' are just references (no need to deallocate them here). */
 typedef struct {
    int SockFD;
    uint_t port;            /* need a separate port in order to support PROXY */
-   bool_t use_proxy;       /* indicates whether to use proxy or not */
+   uint_t flags;
    DilloWeb *web;          /* reference to client's web structure */
    Dlist *addr_list;       /* Holds the DNS answer */
    int Err;                /* Holds the errno of the connect() call */
    ChainLink *Info;        /* Used for CCC asynchronous operations */
+   char *connected_to;     /* Used for per-host connection limit */
 } SocketData_t;
 
+/* Data structures and functions to queue sockets that need to be
+ * delayed due to the per host connection limit.
+ */
+typedef struct SocketQueueEntry {
+   SocketData_t* sock;
+   struct SocketQueueEntry *next ;
+} SocketQueueEntry_t;
+
+typedef struct {
+   SocketQueueEntry_t *head;
+   SocketQueueEntry_t *tail;
+} SocketQueue_t;
+
+typedef struct {
+  char *host;
+  int active_connections;
+  SocketQueue_t queue;
+} HostConnection_t;
+
+static void Http_socket_queue_init(SocketQueue_t *sq);
+static void Http_socket_enqueue(SocketQueue_t *sq, SocketData_t* sock);
+static SocketData_t* Http_socket_dequeue(SocketQueue_t *sq);
+static HostConnection_t *Http_host_connection_get(const char *host);
+static void Http_host_connection_remove(HostConnection_t *hc);
+static int Http_connect_socket(ChainLink *Info);
+static void Http_socket_free(int SKey);
 
 /*
  * Local data
  */
 static Klist_t *ValidSocks = NULL; /* Active sockets list. It holds pointers to
                                     * SocketData_t structures. */
-
 static DilloUrl *HTTP_Proxy = NULL;
 static char *HTTP_Proxy_Auth_base64 = NULL;
 static char *HTTP_Language_hdr = NULL;
+static Dlist *host_connections;
 
 /*
  * Initialize proxy vars and Accept-Language header
@@ -91,6 +123,9 @@ int a_Http_init(void)
    if (HTTP_Proxy && prefs.http_proxyuser && strchr(prefs.http_proxyuser, ':'))
       HTTP_Proxy_Auth_base64 = a_Misc_encode_base64(prefs.http_proxyuser);
  */
+
+   host_connections = dList_new(5);
+
    return 0;
 }
 
@@ -123,6 +158,31 @@ static int Http_sock_new(void)
    return a_Klist_insert(&ValidSocks, S);
 }
 
+static void Http_connect_queued_sockets(HostConnection_t *hc)
+{
+   SocketData_t *sd;
+   while (hc->active_connections < prefs.http_max_conns &&
+          (sd = Http_socket_dequeue(&hc->queue))) {
+
+      sd->flags &= ~HTTP_SOCKET_QUEUED;
+
+      if (sd->flags & HTTP_SOCKET_TO_BE_FREED) {
+          dFree(sd);
+      } else if (a_Web_valid(sd->web)) {
+         /* start connecting the socket */
+         if (Http_connect_socket(sd->Info) < 0) {
+            MSG_BW(sd->web, 1, "ERROR: %s", dStrerror(sd->Err));
+            a_Chain_bfcb(OpAbort, sd->Info, NULL, "Both");
+            dFree(sd->Info);
+            Http_socket_free(VOIDP2INT(sd->Info->LocalKey));
+         } else {
+            sd->connected_to = hc->host;
+            hc->active_connections++;
+         }
+      }
+   }
+}
+
 /*
  * Free SocketData_t struct
  */
@@ -132,7 +192,19 @@ static void Http_socket_free(int SKey)
 
    if ((S = a_Klist_get_data(ValidSocks, SKey))) {
       a_Klist_remove(ValidSocks, SKey);
-      dFree(S);
+
+      if (S->flags & HTTP_SOCKET_QUEUED) {
+         S->flags |= HTTP_SOCKET_TO_BE_FREED;
+      } else {
+         if (S->connected_to) {
+            HostConnection_t *hc = Http_host_connection_get(S->connected_to);
+            hc->active_connections--;
+            Http_connect_queued_sockets(hc);
+            if (hc->active_connections == 0)
+               Http_host_connection_remove(hc);
+      }
+         dFree(S);
+      }
    }
 }
 
@@ -233,6 +305,7 @@ Dstr *a_Http_make_query_str(const DilloUrl *url, bool_t use_proxy)
          query,
          "POST %s HTTP/1.1\r\n"
          "Connection: close\r\n"
+         "Accept: text/*,image/*,*/*;q=0.2\r\n"
          "Accept-Charset: utf-8,*;q=0.8\r\n"
          "Accept-Encoding: gzip\r\n"
          "%s" /* language */
@@ -240,13 +313,13 @@ Dstr *a_Http_make_query_str(const DilloUrl *url, bool_t use_proxy)
          "Host: %s\r\n"
          "%s"
          "%s"
-         "User-Agent: Dillo/%s\r\n"
+         "User-Agent: %s\r\n"
          "Content-Length: %ld\r\n"
          "Content-Type: %s\r\n"
          "%s" /* cookies */
          "\r\n",
          full_path->str, HTTP_Language_hdr, auth ? auth : "",
-         URL_AUTHORITY(url), proxy_auth->str, referer, VERSION,
+         URL_AUTHORITY(url), proxy_auth->str, referer, prefs.http_user_agent,
          (long)URL_DATA(url)->len, content_type->str,
          cookies);
       dStr_append_l(query, URL_DATA(url)->str, URL_DATA(url)->len);
@@ -257,6 +330,7 @@ Dstr *a_Http_make_query_str(const DilloUrl *url, bool_t use_proxy)
          "GET %s HTTP/1.1\r\n"
          "%s"
          "Connection: close\r\n"
+         "Accept: text/*,image/*,*/*;q=0.2\r\n"
          "Accept-Charset: utf-8,*;q=0.8\r\n"
          "Accept-Encoding: gzip\r\n"
          "%s" /* language */
@@ -264,14 +338,14 @@ Dstr *a_Http_make_query_str(const DilloUrl *url, bool_t use_proxy)
          "Host: %s\r\n"
          "%s"
          "%s"
-         "User-Agent: Dillo/%s\r\n"
+         "User-Agent: %s\r\n"
          "%s" /* cookies */
          "\r\n",
          full_path->str,
          (URL_FLAGS(url) & URL_E2EQuery) ?
             "Cache-Control: no-cache\r\nPragma: no-cache\r\n" : "",
          HTTP_Language_hdr, auth ? auth : "", URL_AUTHORITY(url),
-         proxy_auth->str, referer, VERSION, cookies);
+         proxy_auth->str, referer, prefs.http_user_agent, cookies);
    }
    dFree(referer);
    dFree(cookies);
@@ -291,7 +365,7 @@ static void Http_send_query(ChainLink *Info, SocketData_t *S)
    DataBuf *dbuf;
 
    /* Create the query */
-   query = a_Http_make_query_str(S->web->url, S->use_proxy);
+   query = a_Http_make_query_str(S->web->url,S->flags & HTTP_SOCKET_USE_PROXY);
    dbuf = a_Chain_dbuf_new(query->str, query->len, 0);
 
    /* actually this message is sent too early.
@@ -299,14 +373,9 @@ static void Http_send_query(ChainLink *Info, SocketData_t *S)
    _MSG_BW(S->web, 1, "Sending query to %s...", URL_HOST_(S->web->url));
 
    /* send query */
-   a_Chain_link_new(Info, a_Http_ccc, BCK, a_IO_ccc, 1, 1);
-   a_Chain_bcb(OpStart, Info, &S->SockFD, NULL);
    a_Chain_bcb(OpSend, Info, dbuf, NULL);
    dFree(dbuf);
    dStr_free(query, 1);
-
-   /* Tell the cache to start the receiving CCC for the answer */
-   a_Chain_fcb(OpSend, Info, &S->SockFD, "SockFD");
 }
 
 /*
@@ -379,6 +448,8 @@ static int Http_connect_socket(ChainLink *Info)
          Http_socket_close(S);
          MSG("Http_connect_socket ERROR: %s\n", dStrerror(S->Err));
       } else {
+         a_Chain_bcb(OpSend, Info, &S->SockFD, "FD");
+         a_Chain_fcb(OpSend, Info, &S->SockFD, "FD");
          Http_send_query(S->Info, S);
          return 0; /* Success */
       }
@@ -399,9 +470,15 @@ static int Http_must_use_proxy(const DilloUrl *url)
    if (HTTP_Proxy) {
       ret = 1;
       if (prefs.no_proxy) {
+         const char *host = URL_HOST(url);
+         size_t host_len = strlen(host);
+
          np = dStrdup(prefs.no_proxy);
          for (p = np; (tok = dStrsep(&p, " "));  ) {
-            if (dStristr(URL_AUTHORITY(url), tok)) {
+            int start = host_len - strlen(tok);
+
+            if (start >= 0 && dStrcasecmp(host + start, tok) == 0) {
+               /* no_proxy token is suffix of host string */
                ret = 0;
                break;
             }
@@ -444,7 +521,7 @@ char *a_Http_make_connect_str(const DilloUrl *url)
       "Host: %s\r\n"
       "%s"
       "\r\n",
-      auth2, 
+      auth2,
       auth2,
       proxy_auth);
 
@@ -468,34 +545,35 @@ const char *a_Http_get_proxy_urlstr()
  * Continue connecting the socket, or abort upon error condition.
  * S->web is checked to assert the operation wasn't aborted while waiting.
  */
-void a_Http_dns_cb(int Status, Dlist *addr_list, void *data)
+static void Http_dns_cb(int Status, Dlist *addr_list, void *data)
 {
    int SKey = VOIDP2INT(data);
    SocketData_t *S;
+   HostConnection_t *hc;
 
    S = a_Klist_get_data(ValidSocks, SKey);
    if (S) {
       if (!a_Web_valid(S->web)) {
-         a_Chain_fcb(OpAbort, S->Info, NULL, NULL);
+         a_Chain_bfcb(OpAbort, S->Info, NULL, "Both");
          dFree(S->Info);
          Http_socket_free(SKey);
 
       } else if (Status == 0 && addr_list) {
          /* Successful DNS answer; save the IP */
          S->addr_list = addr_list;
-         /* start connecting the socket */
-         if (Http_connect_socket(S->Info) < 0) {
-            MSG_BW(S->web, 1, "ERROR: %s", dStrerror(S->Err));
-            a_Chain_fcb(OpAbort, S->Info, NULL, NULL);
-            dFree(S->Info);
-            Http_socket_free(SKey);
-         }
-
+         S->flags |= HTTP_SOCKET_QUEUED;
+         if (S->flags & HTTP_SOCKET_USE_PROXY)
+            hc = Http_host_connection_get(URL_HOST(HTTP_Proxy));
+         else
+            hc = Http_host_connection_get(URL_HOST(S->web->url));
+         Http_socket_enqueue(&hc->queue, S);
+         Http_connect_queued_sockets(hc);
       } else {
          /* DNS wasn't able to resolve the hostname */
          MSG_BW(S->web, 0, "ERROR: Dns can't resolve %s",
-            (S->use_proxy) ? URL_HOST_(HTTP_Proxy) : URL_HOST_(S->web->url));
-         a_Chain_fcb(OpAbort, S->Info, NULL, NULL);
+            (S->flags & HTTP_SOCKET_USE_PROXY) ? URL_HOST_(HTTP_Proxy) :
+                                                 URL_HOST_(S->web->url));
+         a_Chain_bfcb(OpAbort, S->Info, NULL, "Both");
          dFree(S->Info);
          Http_socket_free(SKey);
       }
@@ -524,11 +602,11 @@ static int Http_get(ChainLink *Info, void *Data1)
    if (Http_must_use_proxy(S->web->url)) {
       hostname = dStrdup(URL_HOST(HTTP_Proxy));
       S->port = URL_PORT(HTTP_Proxy);
-      S->use_proxy = TRUE;
+      S->flags |= HTTP_SOCKET_USE_PROXY;
    } else {
       hostname = dStrdup(URL_HOST(S->web->url));
       S->port = URL_PORT(S->web->url);
-      S->use_proxy = FALSE;
+      S->flags &= ~HTTP_SOCKET_USE_PROXY;
    }
 
    /* Let the user know what we'll do */
@@ -536,7 +614,7 @@ static int Http_get(ChainLink *Info, void *Data1)
 
    /* Let the DNS engine resolve the hostname, and when done,
     * we'll try to connect the socket from the callback function */
-   a_Dns_resolve(hostname, a_Http_dns_cb, Info->LocalKey);
+   a_Dns_resolve(hostname, Http_dns_cb, Info->LocalKey);
 
    dFree(hostname);
    return 0;
@@ -550,6 +628,8 @@ void a_Http_ccc(int Op, int Branch, int Dir, ChainLink *Info,
 {
    int SKey = VOIDP2INT(Info->LocalKey);
 
+   (void)Data2; /* suppress unused parameter warning */
+
    dReturn_if_fail( a_Chain_check("a_Http_ccc", Op, Branch, Dir, Info) );
 
    if (Branch == 1) {
@@ -560,6 +640,10 @@ void a_Http_ccc(int Op, int Branch, int Dir, ChainLink *Info,
             /* ( Data1 = Web ) */
             SKey = Http_sock_new();
             Info->LocalKey = INT2VOIDP(SKey);
+            /* link IO */
+            a_Chain_link_new(Info, a_Http_ccc, BCK, a_IO_ccc, 1, 1);
+            a_Chain_bcb(OpStart, Info, NULL, NULL);
+            /* async. connection */
             Http_get(Info, Data1);
             break;
          case OpEnd:
@@ -575,7 +659,7 @@ void a_Http_ccc(int Op, int Branch, int Dir, ChainLink *Info,
             dFree(Info);
             break;
          }
-      } else {  /* FWD */
+      } else {  /* 1 FWD */
          /* HTTP send-query status branch */
          switch (Op) {
          default:
@@ -587,6 +671,82 @@ void a_Http_ccc(int Op, int Branch, int Dir, ChainLink *Info,
 }
 
 
+static void Http_socket_queue_init(SocketQueue_t *sq)
+{
+   sq->head = NULL;
+   sq->tail = NULL;
+}
+
+static void Http_socket_enqueue(SocketQueue_t *sq, SocketData_t* sock)
+{
+   SocketQueueEntry_t *se = dNew(SocketQueueEntry_t, 1);
+
+   se->sock = sock;
+   se->next = NULL;
+
+   if (sq->tail)
+      sq->tail->next = se;
+   sq->tail = se;
+
+   if (! sq->head)
+      sq->head = se;
+}
+
+static SocketData_t* Http_socket_dequeue(SocketQueue_t *sq)
+{
+   SocketQueueEntry_t *se = sq->head;
+   SocketData_t *sd = NULL;
+
+   if (se) {
+      sq->head = se->next;
+      if (sq->tail == se)
+         sq->tail = NULL;
+      sd = se->sock;
+      dFree(se);
+   }
+
+   return sd;
+}
+
+static HostConnection_t *Http_host_connection_get(const char *host)
+{
+   int i;
+   HostConnection_t *hc;
+
+   for (i = 0; i < dList_length(host_connections); i++) {
+      hc = (HostConnection_t*) dList_nth_data(host_connections, i);
+
+      if (dStrcasecmp(host, hc->host) == 0)
+         return hc;
+   }
+
+   hc = dNew0(HostConnection_t, 1);
+   Http_socket_queue_init(&hc->queue);
+   hc->host = dStrdup(host);
+   dList_append(host_connections, hc);
+
+   return hc;
+}
+
+static void Http_host_connection_remove(HostConnection_t *hc)
+{
+    assert(hc->queue.head == NULL);
+    dList_remove_fast(host_connections, hc);
+    dFree(hc->host);
+    dFree(hc);
+}
+
+static void Http_host_connection_remove_all()
+{
+   HostConnection_t *hc;
+
+   while (dList_length(host_connections) > 0) {
+      hc = (HostConnection_t*) dList_nth_data(host_connections, 0);
+      while (Http_socket_dequeue(&hc->queue));
+      Http_host_connection_remove(hc);
+   }
+   dList_free(host_connections);
+}
 
 /*
  * Deallocate memory used by http module
@@ -594,6 +754,7 @@ void a_Http_ccc(int Op, int Branch, int Dir, ChainLink *Info,
  */
 void a_Http_freeall(void)
 {
+   Http_host_connection_remove_all();
    a_Klist_free(&ValidSocks);
    a_Url_free(HTTP_Proxy);
    dFree(HTTP_Proxy_Auth_base64);
