@@ -85,10 +85,10 @@ static uint_t DelayedQueueIdleId = 0;
 /*
  *  Forward declarations
  */
-static void Cache_process_queue(CacheEntry_t *entry);
+static CacheEntry_t *Cache_process_queue(CacheEntry_t *entry);
 static void Cache_delayed_process_queue(CacheEntry_t *entry);
 static void Cache_auth_entry(CacheEntry_t *entry, BrowserWindow *bw);
-
+static void Cache_entry_inject(const DilloUrl *Url, Dstr *data_ds);
 
 /*
  * Determine if two cache entries are equal (used by CachedURLs)
@@ -124,7 +124,7 @@ void a_Cache_init(void)
    {
       DilloUrl *url = a_Url_new("about:splash", NULL);
       Dstr *ds = dStr_new(AboutSplash);
-      a_Cache_entry_inject(url, ds);
+      Cache_entry_inject(url, ds);
       dStr_free(ds, 1);
       a_Url_free(url);
    }
@@ -151,6 +151,7 @@ static int Cache_client_enqueue(const DilloUrl *Url, DilloWeb *Web,
    NewClient->Url = Url;
    NewClient->Version = 0;
    NewClient->Buf = NULL;
+   NewClient->BufSize = 0;
    NewClient->Callback = Callback;
    NewClient->CbData = CbData;
    NewClient->Web    = Web;
@@ -266,7 +267,7 @@ static CacheEntry_t *Cache_entry_add(const DilloUrl *Url)
  * Inject full page content directly into the cache.
  * Used for "about:splash". May be used for "about:cache" too.
  */
-void a_Cache_entry_inject(const DilloUrl *Url, Dstr *data_ds)
+static void Cache_entry_inject(const DilloUrl *Url, Dstr *data_ds)
 {
    CacheEntry_t *entry;
 
@@ -310,6 +311,10 @@ static void Cache_entry_free(CacheEntry_t *entry)
    dStr_free(entry->UTF8Data, 1);
    if (entry->CharsetDecoder)
       a_Decode_free(entry->CharsetDecoder);
+   if (entry->TransferDecoder)
+      a_Decode_free(entry->TransferDecoder);
+   if (entry->ContentDecoder)
+      a_Decode_free(entry->ContentDecoder);
    dFree(entry);
 }
 
@@ -490,7 +495,7 @@ const char *a_Cache_set_content_type(const DilloUrl *url, const char *ctype,
 {
    const char *curr;
    char *major, *minor, *charset;
-   CacheEntry_t *entry = Cache_entry_search_with_redirect(url);
+   CacheEntry_t *entry = Cache_entry_search(url);
 
    dReturn_val_if_fail (entry != NULL, NULL);
 
@@ -515,6 +520,10 @@ const char *a_Cache_set_content_type(const DilloUrl *url, const char *ctype,
              ((!major || !*major) && (!minor || !*minor))) {
             /* META only gives charset; use detected MIME type too */
             entry->TypeNorm = dStrconcat(entry->TypeDet, ctype, NULL);
+         } else if (*from == 'm' &&
+                    !dStrncasecmp(ctype, "text/xhtml", 10)) {
+            /* WORKAROUND: doxygen uses "text/xhtml" in META */
+            entry->TypeNorm = dStrdup(entry->TypeDet);
          }
          if (charset) {
             if (entry->CharsetDecoder)
@@ -584,14 +593,16 @@ static char *Cache_parse_field(const char *header, const char *fieldname)
       }
 
       i += j;
-      while (header[i] == ' ') i++;
       if (header[i] == ':') {
         /* Field found! */
         while (header[++i] == ' ' || header[i] == '\t');
         for (j = 0; header[i + j] != '\n'; j++);
+        while (j && (header[i + j - 1] == ' ' || header[i + j - 1] == '\t'))
+           j--;
         field = dStrndup(header + i, j);
         return field;
       }
+      while (header[i] != '\n') i++;
    }
    return NULL;
 }
@@ -618,13 +629,16 @@ static Dlist *Cache_parse_multiple_fields(const char *header,
       }
 
       i += j;
-      for ( ; header[i] == ' '; i++);
       if (header[i] == ':') {
          /* Field found! */
          while (header[++i] == ' ' || header[i] == '\t');
          for (j = 0; header[i + j] != '\n'; j++);
+         while (j && (header[i + j - 1] == ' ' || header[i + j - 1] == '\t'))
+            j--;
          field = dStrndup(header + i, j);
          dList_append(fields, field);
+      } else {
+         while (header[i] != '\n') i++;
       }
    }
 
@@ -701,16 +715,22 @@ static void Cache_parse_header(CacheEntry_t *entry)
       dList_free(warnings);
    }
 
+   /*
+    * Get Transfer-Encoding and initialize decoder
+    */
+   encoding = Cache_parse_field(header, "Transfer-Encoding");
+   entry->TransferDecoder = a_Decode_transfer_init(encoding);
+
+
    if ((Length = Cache_parse_field(header, "Content-Length")) != NULL) {
-      char *tmp;
-      if ((tmp = Cache_parse_field(header, "Transfer-Encoding"))) {
+      if (encoding) {
          /*
-          * BUG: Should test for _presence_ of headers, not whether they
-          * have content.
+          * If Transfer-Encoding is present, Content-Length must be ignored.
+          * If the Transfer-Encoding is non-identity, it is an error.
           */
-         MSG_HTTP("Both Content-Length and Transfer-Encoding headers"
-                  " received.\n");
-         dFree(tmp);
+         if (dStrcasecmp(encoding, "identity"))
+            MSG_HTTP("Content-Length and non-identity Transfer-Encoding "
+                     "headers both present.\n");
       } else {
          entry->Flags |= CA_GotLength;
          entry->ExpectedSize = MAX(strtol(Length, NULL, 10), 0);
@@ -718,26 +738,20 @@ static void Cache_parse_header(CacheEntry_t *entry)
       dFree(Length);
    }
 
+   dFree(encoding); /* free Transfer-Encoding */
+
 #ifndef DISABLE_COOKIES
-   /* BUG: If a server feels like mixing Set-Cookie2 and Set-Cookie
-    * responses which aren't identical, then we have a problem. I don't
-    * know if that is a real issue though. */
-   if ((Cookies = Cache_parse_multiple_fields(header, "Set-Cookie2")) ||
-       (Cookies = Cache_parse_multiple_fields(header, "Set-Cookie"))) {
-      a_Cookies_set(Cookies, entry->Url);
+   if ((Cookies = Cache_parse_multiple_fields(header, "Set-Cookie"))) {
+      char *server_date = Cache_parse_field(header, "Date");
+
+      a_Cookies_set(Cookies, entry->Url, server_date);
       for (i = 0; (data = dList_nth_data(Cookies, i)); ++i)
          dFree(data);
       dList_free(Cookies);
+      dFree(server_date);
    }
 #endif /* !DISABLE_COOKIES */
 
-   /*
-    * Get Transfer-Encoding and initialize decoder
-    */
-   encoding = Cache_parse_field(header, "Transfer-Encoding");
-   entry->TransferDecoder = a_Decode_transfer_init(encoding);
-   dFree(encoding);
-
    /*
     * Get Content-Encoding and initialize decoder
     */
@@ -757,12 +771,7 @@ static void Cache_parse_header(CacheEntry_t *entry)
    }
 
    /* Get Content-Type */
-   if ((Type = Cache_parse_field(header, "Content-Type")) == NULL) {
-      if (!((entry->Flags & CA_GotLength) && (entry->ExpectedSize == 0))) {
-         /* unless the server sent Content-Length: 0 */
-         MSG_HTTP("Server didn't send Content-Type in header.\n");
-      }
-   } else {
+   if ((Type = Cache_parse_field(header, "Content-Type"))) {
       /* This HTTP Content-Type is not trusted. It's checked against real data
        * in Cache_process_queue(); only then CA_GotContentType becomes true. */
       a_Cache_set_content_type(entry->Url, Type, "http");
@@ -870,9 +879,13 @@ void a_Cache_process_dbuf(int Op, const char *buf, size_t buf_size,
          if (entry->Data->len)
             entry->Flags &= ~CA_IsEmpty;
 
-         Cache_process_queue(entry);
+         entry = Cache_process_queue(entry);
       }
    } else if (Op == IOClose) {
+      if ((entry->ExpectedSize || entry->TransferSize) &&
+          entry->TypeHdr == NULL) {
+         MSG_HTTP("Message with a body lacked Content-Type header.\n");
+      }
       if ((entry->Flags & CA_GotLength) &&
           (entry->ExpectedSize != entry->TransferSize)) {
          MSG_HTTP("Content-Length does NOT match message body,\n"
@@ -880,6 +893,16 @@ void a_Cache_process_dbuf(int Op, const char *buf, size_t buf_size,
          MSG("entry->ExpectedSize = %d, entry->TransferSize = %d\n",
              entry->ExpectedSize, entry->TransferSize);
       }
+      if (!entry->TransferSize && !(entry->Flags & CA_Redirect) &&
+          (entry->Flags & WEB_RootUrl)) {
+         char *eol = strchr(entry->Header->str, '\n');
+         if (eol) {
+            char *status_line = dStrndup(entry->Header->str,
+                                         eol - entry->Header->str);
+            MSG_HTTP("Body was empty. Server sent status: %s\n", status_line);
+            dFree(status_line);
+         }
+      }
       entry->Flags |= CA_GotData;
       entry->Flags &= ~CA_Stopped;          /* it may catch up! */
       if (entry->TransferDecoder) {
@@ -891,11 +914,12 @@ void a_Cache_process_dbuf(int Op, const char *buf, size_t buf_size,
          entry->ContentDecoder = NULL;
       }
       dStr_fit(entry->Data);                /* fit buffer size! */
-      Cache_process_queue(entry);
-      if (entry->Flags & CA_GotHeader) {
-         Cache_unref_data(entry);
-      }
 
+      if ((entry = Cache_process_queue(entry))) {
+         if (entry->Flags & CA_GotHeader) {
+            Cache_unref_data(entry);
+         }
+      }
    } else if (Op == IOAbort) {
       /* unused */
       MSG("a_Cache_process_dbuf Op = IOAbort; not implemented!\n");
@@ -926,7 +950,7 @@ static int Cache_redirect(CacheEntry_t *entry, int Flags, BrowserWindow *bw)
        (entry->Flags & CA_ForceRedirect || entry->Flags & CA_TempRedirect ||
         !entry->Data->len || entry->Data->len < 1024)) {
 
-      _MSG(">>>Redirect from: %s\n to %s\n",
+      _MSG(">>>> Redirect from: %s\n to %s <<<<\n",
            URL_STR_(entry->Url), URL_STR_(entry->Location));
       _MSG("%s", entry->Header->str);
 
@@ -935,14 +959,14 @@ static int Cache_redirect(CacheEntry_t *entry, int Flags, BrowserWindow *bw)
          NewUrl = a_Url_new(URL_STR_(entry->Location), URL_STR_(entry->Url));
          if (entry->Flags & CA_TempRedirect)
             a_Url_set_flags(NewUrl, URL_FLAGS(NewUrl) | URL_E2EQuery);
-         a_Nav_push(bw, NewUrl);
+         a_Nav_push(bw, NewUrl, entry->Url);
          a_Url_free(NewUrl);
       } else {
          /* Sub entity redirection (most probably an image) */
          if (!entry->Data->len) {
-            _MSG(">>>Image redirection without entity-content<<<\n");
+            _MSG(">>>> Image redirection without entity-content <<<<\n");
          } else {
-            _MSG(">>>Image redirection with entity-content<<<\n");
+            _MSG(">>>> Image redirection with entity-content <<<<\n");
          }
       }
    }
@@ -997,7 +1021,7 @@ static void Cache_auth_entry(CacheEntry_t *entry, BrowserWindow *bw)
  * Check whether a URL scheme is downloadable.
  * Return: 1 enabled, 0 disabled.
  */
-static int Cache_download_enabled(const DilloUrl *url)
+int a_Cache_download_enabled(const DilloUrl *url)
 {
    if (!dStrcasecmp(URL_SCHEME(url), "http") ||
        !dStrcasecmp(URL_SCHEME(url), "https") ||
@@ -1028,6 +1052,24 @@ static void Cache_null_client(int Op, CacheClient_t *Client)
    return;
 }
 
+typedef struct {
+   BrowserWindow *bw;
+   DilloUrl *url;
+} Cache_savelink_t;
+
+/*
+ * Save link from behind a timeout so that Cache_process_queue() can
+ * get on with its work.
+ */
+static void Cache_savelink_cb(void *vdata)
+{
+   Cache_savelink_t *data = (Cache_savelink_t*) vdata;
+
+   a_UIcmd_save_link(data->bw, data->url);
+   a_Url_free(data->url);
+   dFree(data);
+}
+
 /*
  * Update cache clients for a single cache-entry
  * Tasks:
@@ -1036,9 +1078,11 @@ static void Cache_null_client(int Op, CacheClient_t *Client)
  *   - Remove clients when done
  *   - Call redirect handler
  *
+ * Return: Cache entry, which may be NULL if it has been removed.
+ *
  * TODO: Implement CA_Abort Op in client callback
  */
-static void Cache_process_queue(CacheEntry_t *entry)
+static CacheEntry_t *Cache_process_queue(CacheEntry_t *entry)
 {
    uint_t i;
    int st;
@@ -1055,7 +1099,7 @@ static void Cache_process_queue(CacheEntry_t *entry)
    if (Busy)
       MSG_ERR("FATAL!: >>>> Cache_process_queue Caught busy!!! <<<<\n");
    if (!(entry->Flags & CA_GotHeader))
-      return;
+      return entry;
    if (!(entry->Flags & CA_GotContentType)) {
       st = a_Misc_get_content_type_from_data(
               entry->Data->str, entry->Data->len, &Type);
@@ -1069,7 +1113,7 @@ static void Cache_process_queue(CacheEntry_t *entry)
          entry->TypeDet = dStrdup(Type);
          entry->Flags |= CA_GotContentType;
       } else
-         return;  /* i.e., wait for more data */
+         return entry;  /* i.e., wait for more data */
    }
 
    Busy = TRUE;
@@ -1132,6 +1176,8 @@ static void Cache_process_queue(CacheEntry_t *entry)
                }
             }
             if (AbortEntry) {
+               if (ClientWeb->flags & WEB_RootUrl)
+                  a_Nav_cancel_expect_if_eq(Client_bw, Client->Url);
                a_Bw_remove_client(Client_bw, Client->Key);
                Cache_client_dequeue(Client, NULLKey);
                --i; /* Keep the index value in the next iteration */
@@ -1177,12 +1223,18 @@ static void Cache_process_queue(CacheEntry_t *entry)
       /* Abort the entry, remove it from cache, and maybe offer download. */
       DilloUrl *url = a_Url_dup(entry->Url);
       a_Capi_conn_abort_by_url(url);
-      Cache_entry_remove(entry, NULL);
-      if (OfferDownload && Cache_download_enabled(url)) {
-         a_UIcmd_save_link(Client_bw, url);
+      entry = NULL;
+      if (OfferDownload) {
+         /* Remove entry when 'conn' is already done */
+         Cache_entry_remove(NULL, url);
+         if (a_Cache_download_enabled(url)) {
+            Cache_savelink_t *data = dNew(Cache_savelink_t, 1);
+            data->bw = Client_bw;
+            data->url = a_Url_dup(url);
+            a_Timeout_add(0.0, Cache_savelink_cb, data);
+         }
       }
       a_Url_free(url);
-
    } else if (entry->Auth && (entry->Flags & CA_GotData)) {
       Cache_auth_entry(entry, Client_bw);
    }
@@ -1195,6 +1247,7 @@ static void Cache_process_queue(CacheEntry_t *entry)
 
    Busy = FALSE;
    _MSG("QueueSize ====> %d\n", dList_length(ClientQueue));
+   return entry;
 }
 
 /*
@@ -1206,10 +1259,7 @@ static void Cache_delayed_process_queue_callback()
 
    while ((entry = (CacheEntry_t *)dList_nth_data(DelayedQueue, 0))) {
       Cache_ref_data(entry);
-      Cache_process_queue(entry);
-      if (entry != dList_nth_data(DelayedQueue, 0)) {
-         /* Cache_process_queue() has removed the entry! */
-      } else {
+      if ((entry = Cache_process_queue(entry))) {
          Cache_unref_data(entry);
          dList_remove(DelayedQueue, entry);
       }
@@ -1299,7 +1349,7 @@ void a_Cache_freeall(void)
 
    /* Remove every cache entry */
    while ((data = dList_nth_data(CachedURLs, 0))) {
-      dList_remove(CachedURLs, data);
+      dList_remove_fast(CachedURLs, data);
       Cache_entry_free(data);
    }
    /* Remove the cache list */
