@@ -13,7 +13,16 @@
  * Non blocking pthread-handled Dns scheme
  */
 
-#include <pthread.h>
+
+/*
+ * Uncomment the following line for debugging or gprof profiling.
+ */
+/* #undef D_DNS_THREADED */
+
+#ifdef D_DNS_THREADED
+#  include <pthread.h>
+#endif
+
 
 #include <netdb.h>
 #include <sys/types.h>
@@ -21,6 +30,7 @@
 #include <arpa/inet.h>
 #include <netinet/in.h>
 #include <errno.h>
+#include <unistd.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -30,11 +40,6 @@
 #include "list.h"
 #include "timeout.hh"
 
-/*
- * Uncomment the following line for debugging or gprof profiling.
- */
-/* #undef D_DNS_THREADED */
-
 
 /* Maximum dns resolving threads */
 #ifdef D_DNS_THREADED
@@ -83,7 +88,6 @@ static GDnsCache *dns_cache;
 static int dns_cache_size, dns_cache_size_max;
 static GDnsQueue *dns_queue;
 static int dns_queue_size, dns_queue_size_max;
-static bool_t ipv6_enabled;
 
 
 /* ----------------------------------------------------------------------
@@ -196,16 +200,14 @@ void a_Dns_init(void)
 #endif
    }
 
-   /* IPv6 test */
-   ipv6_enabled = FALSE;
 #ifdef ENABLE_IPV6
+   /* IPv6 test */
    {
       /* If the IPv6 address family is not available there is no point
          wasting time trying to connect to v6 addresses. */
       int fd = socket(AF_INET6, SOCK_STREAM, 0);
       if (fd >= 0) {
          close(fd);
-         ipv6_enabled = TRUE;
       }
    }
 #endif
@@ -481,14 +483,17 @@ static void Dns_timeout_client(void *data)
  *  Dns memory-deallocation
  *  (Call this one at exit time)
  *  The Dns_queue is deallocated at execution time (no need to do that here)
- *  'dns_cache' is the only one that grows dinamically
+ *  'dns_cache' is the only one that grows dynamically
  */
 void a_Dns_freeall(void)
 {
-   int i;
+   int i, j;
 
    for ( i = 0; i < dns_cache_size; ++i ){
       dFree(dns_cache[i].hostname);
+      for ( j = 0; j < dList_length(dns_cache[i].addr_list); ++j)
+         dFree(dList_nth_data(dns_cache[i].addr_list, j));
+      dList_free(dns_cache[i].addr_list);
    }
    dFree(dns_cache);
 }
