@@ -191,16 +191,16 @@ static void Nav_stack_clean(BrowserWindow *bw)
  * This function requests the page's root-URL; images and related stuff
  * are fetched directly by the HTML module.
  */
-static void Nav_open_url(BrowserWindow *bw, const DilloUrl *url, int offset)
+static void Nav_open_url(BrowserWindow *bw, const DilloUrl *url,
+                         const DilloUrl *requester, int offset)
 {
-   DilloUrl *old_url;
-   bool_t MustLoad, ForceReload, Repush, IgnoreScroll;
+   const DilloUrl *old_url;
+   bool_t MustLoad, ForceReload, IgnoreScroll;
    int x, y, idx, ClientKey;
    DilloWeb *Web;
 
    MSG("Nav_open_url: new url='%s'\n", URL_STR_(url));
 
-   Repush = (URL_FLAGS(url) & URL_ReloadFromCache) != 0;
    ForceReload = (URL_FLAGS(url) & (URL_E2EQuery + URL_ReloadFromCache)) != 0;
    IgnoreScroll = (URL_FLAGS(url) & URL_IgnoreScroll) != 0;
 
@@ -232,7 +232,7 @@ static void Nav_open_url(BrowserWindow *bw, const DilloUrl *url, int offset)
 
       // a_Menu_pagemarks_new(bw);
 
-      Web = a_Web_new(url);
+      Web = a_Web_new(url, requester);
       Web->bw = bw;
       Web->flags |= WEB_RootUrl;
       if ((ClientKey = a_Capi_open_url(Web, NULL, NULL)) != 0) {
@@ -248,15 +248,23 @@ static void Nav_open_url(BrowserWindow *bw, const DilloUrl *url, int offset)
  */
 void a_Nav_cancel_expect(BrowserWindow *bw)
 {
-   if (bw->nav_expecting) {
-      a_Url_free(bw->nav_expect_url);
-      bw->nav_expect_url = NULL;
-      bw->nav_expecting = FALSE;
+   if (a_Bw_expecting(bw)) {
+      a_Bw_cancel_expect(bw);
+      a_UIcmd_set_buttons_sens(bw);
    }
    if (bw->meta_refresh_status > 0)
       --bw->meta_refresh_status;
 }
 
+/*
+ * Cancel the expect if 'url' matches.
+ */
+void a_Nav_cancel_expect_if_eq(BrowserWindow *bw, const DilloUrl *url)
+{
+   if (a_Url_cmp(url, a_Bw_expected_url(bw)) == 0)
+      a_Nav_cancel_expect(bw);
+}
+
 /*
  * We have an answer! Set things accordingly.
  * This function is called for root URLs only.
@@ -271,8 +279,8 @@ void a_Nav_expect_done(BrowserWindow *bw)
 
    dReturn_if_fail(bw != NULL);
 
-   if (bw->nav_expecting) {
-      url = bw->nav_expect_url;
+   if (a_Bw_expecting(bw)) {
+      url = a_Url_dup(a_Bw_expected_url(bw));
       reload = (URL_FLAGS(url) & URL_ReloadPage);
       repush = (URL_FLAGS(url) & URL_ReloadFromCache);
       e2equery = (URL_FLAGS(url) & URL_E2EQuery);
@@ -283,6 +291,7 @@ void a_Nav_expect_done(BrowserWindow *bw)
       m = URL_E2EQuery|URL_ReloadPage|URL_ReloadFromCache|URL_IgnoreScroll;
       a_Url_set_flags(url, URL_FLAGS(url) & ~m);
       url_idx = a_History_add_url(url);
+      a_Url_free(url);
 
       if (repush) {
          MSG("a_Nav_expect_done: repush!\n");
@@ -309,7 +318,7 @@ void a_Nav_expect_done(BrowserWindow *bw)
    }
 
    if (goto_old_scroll) {
-      /* Scroll to were we were in this page */
+      /* Scroll to where we were in this page */
       Nav_get_scroll_pos(bw, &posx, &posy);
       a_UIcmd_set_scroll_xy(bw, posx, posy);
       _MSG("Nav: expect_done scrolling to x=%d y=%d\n", posx, posy);
@@ -332,19 +341,21 @@ void a_Nav_expect_done(BrowserWindow *bw)
  * - Set bw to expect the URL data
  * - Ask the cache to feed back the requested URL (via Nav_open_url)
  */
-void a_Nav_push(BrowserWindow *bw, const DilloUrl *url)
+void a_Nav_push(BrowserWindow *bw, const DilloUrl *url,
+                                   const DilloUrl *requester)
 {
+   const DilloUrl *e_url;
    dReturn_if_fail (bw != NULL);
 
-   if (bw->nav_expecting && !a_Url_cmp(bw->nav_expect_url, url) &&
-       !strcmp(URL_FRAGMENT(bw->nav_expect_url),URL_FRAGMENT(url))) {
+   e_url = a_Bw_expected_url(bw);
+   if (e_url && !a_Url_cmp(e_url, url) &&
+       !strcmp(URL_FRAGMENT(e_url),URL_FRAGMENT(url))) {
       /* we're already expecting that url (most probably a double-click) */
       return;
    }
    a_Nav_cancel_expect(bw);
-   bw->nav_expect_url = a_Url_dup(url);
-   bw->nav_expecting = TRUE;
-   Nav_open_url(bw, url, 0);
+   a_Bw_expect(bw, url);
+   Nav_open_url(bw, url, requester, 0);
 }
 
 /*
@@ -359,9 +370,8 @@ static void Nav_repush(BrowserWindow *bw)
       url = a_Url_dup(a_History_get_url(NAV_TOP_UIDX(bw)));
       /* Let's make reload be from Cache */
       a_Url_set_flags(url, URL_FLAGS(url) | URL_ReloadFromCache);
-      bw->nav_expect_url = a_Url_dup(url);
-      bw->nav_expecting = TRUE;
-      Nav_open_url(bw, url, 0);
+      a_Bw_expect(bw, url);
+      Nav_open_url(bw, url, NULL, 0);
       a_Url_free(url);
    }
 }
@@ -384,7 +394,7 @@ static void Nav_repush_callback(void *data)
 void a_Nav_repush(BrowserWindow *bw)
 {
    dReturn_if_fail (bw != NULL);
-   MSG(">>> a_Nav_repush <<<<\n");
+   MSG(">>>> a_Nav_repush <<<<\n");
    a_Timeout_add(0.0, Nav_repush_callback, (void*)bw);
 }
 
@@ -394,11 +404,12 @@ void a_Nav_repush(BrowserWindow *bw)
 static void Nav_redirection0_callback(void *data)
 {
    BrowserWindow *bw = (BrowserWindow *)data;
+   const DilloUrl *referer_url = a_History_get_url(NAV_TOP_UIDX(bw));
    _MSG(">>>> Nav_redirection0_callback <<<<\n");
 
    if (bw->meta_refresh_status == 2) {
       Nav_stack_move_ptr(bw, -1);
-      a_Nav_push(bw, bw->meta_refresh_url);
+      a_Nav_push(bw, bw->meta_refresh_url, referer_url);
    }
    a_Url_free(bw->meta_refresh_url);
    bw->meta_refresh_url = NULL;
@@ -412,7 +423,7 @@ static void Nav_redirection0_callback(void *data)
 void a_Nav_redirection0(BrowserWindow *bw, const DilloUrl *new_url)
 {
    dReturn_if_fail (bw != NULL);
-   _MSG(">>> a_Nav_redirection0 <<<<\n");
+   _MSG(">>>> a_Nav_redirection0 <<<<\n");
 
    a_Url_free(bw->meta_refresh_url);
    bw->meta_refresh_url = a_Url_dup(new_url);
@@ -422,27 +433,6 @@ void a_Nav_redirection0(BrowserWindow *bw, const DilloUrl *new_url)
    a_Timeout_add(0.0, Nav_redirection0_callback, (void*)bw);
 }
 
-/*
- * Same as a_Nav_push() but in a new window.
- */
-void a_Nav_push_nw(BrowserWindow *bw, const DilloUrl *url)
-{
-   int w, h;
-   BrowserWindow *newbw;
-
-   a_UIcmd_get_wh(bw, &w, &h);
-   newbw = a_UIcmd_browser_window_new(w, h, 0, bw);
-   a_Nav_push(newbw, url);
-}
-
-/*
- * Wraps a_Nav_push to match 'DwPage->link' function type
- */
-void a_Nav_vpush(void *vbw, const DilloUrl *url)
-{
-   a_Nav_push(vbw, url);
-}
-
 /*
  * Send the browser back to previous page
  */
@@ -453,7 +443,7 @@ void a_Nav_back(BrowserWindow *bw)
    a_Nav_cancel_expect(bw);
    if (--idx >= 0){
       a_UIcmd_set_msg(bw, "");
-      Nav_open_url(bw, a_History_get_url(NAV_UIDX(bw,idx)), -1);
+      Nav_open_url(bw, a_History_get_url(NAV_UIDX(bw,idx)), NULL, -1);
    }
 }
 
@@ -467,7 +457,7 @@ void a_Nav_forw(BrowserWindow *bw)
    a_Nav_cancel_expect(bw);
    if (++idx < a_Nav_stack_size(bw)) {
       a_UIcmd_set_msg(bw, "");
-      Nav_open_url(bw, a_History_get_url(NAV_UIDX(bw,idx)), +1);
+      Nav_open_url(bw, a_History_get_url(NAV_UIDX(bw,idx)), NULL, +1);
    }
 }
 
@@ -476,56 +466,56 @@ void a_Nav_forw(BrowserWindow *bw)
  */
 void a_Nav_home(BrowserWindow *bw)
 {
-   a_Nav_push(bw, prefs.home);
+   a_Nav_push(bw, prefs.home, NULL);
 }
 
 /*
  * This one does a_Nav_reload's job!
  */
-static void Nav_reload(BrowserWindow *bw)
+static void Nav_reload_callback(void *data)
 {
-   DilloUrl *ReqURL;
+   BrowserWindow *bw = data;
+   const DilloUrl *h_url;
+   DilloUrl *r_url;
+   int choice, confirmed = 1;
 
    a_Nav_cancel_expect(bw);
    if (a_Nav_stack_size(bw)) {
-      ReqURL = a_Url_dup(a_History_get_url(NAV_TOP_UIDX(bw)));
-      /* Mark URL as reload to differentiate from push */
-      a_Url_set_flags(ReqURL, URL_FLAGS(ReqURL) | URL_ReloadPage);
-      /* Let's make reload be end-to-end */
-      a_Url_set_flags(ReqURL, URL_FLAGS(ReqURL) | URL_E2EQuery);
-      /* This is an explicit reload, so clear the SpamSafe flag */
-      a_Url_set_flags(ReqURL, URL_FLAGS(ReqURL) & ~URL_SpamSafe);
-      bw->nav_expect_url = ReqURL;
-      bw->nav_expecting = TRUE;
-      Nav_open_url(bw, ReqURL, 0);
+      h_url = a_History_get_url(NAV_TOP_UIDX(bw));
+      if (strncmp(URL_STR(h_url), "dpi:/vsource/", 13) == 0) {
+         /* disable reload for view source dpi */
+         confirmed = 0;
+      } else if (URL_FLAGS(h_url) & URL_Post) {
+         /* Attempt to repost data, let's confirm... */
+         choice = a_Dialog_choice5("Repost form data?",
+                                   "No", "Yes", "Cancel", NULL, NULL);
+         confirmed = (choice == 2);  /* "Yes" */
+      }
+
+      if (confirmed) {
+         r_url = a_Url_dup(h_url);
+         /* Mark URL as reload to differentiate from push */
+         a_Url_set_flags(r_url, URL_FLAGS(r_url) | URL_ReloadPage);
+         /* Let's make reload be end-to-end */
+         a_Url_set_flags(r_url, URL_FLAGS(r_url) | URL_E2EQuery);
+         /* This is an explicit reload, so clear the SpamSafe flag */
+         a_Url_set_flags(r_url, URL_FLAGS(r_url) & ~URL_SpamSafe);
+         a_Bw_expect(bw, r_url);
+         Nav_open_url(bw, r_url, NULL, 0);
+         a_Url_free(r_url);
+      }
    }
 }
 
 /*
  * Implement the RELOAD button functionality.
  * (Currently it only reloads the page, not its images)
+ * Note: the timeout lets CCC operations end before making the request.
  */
 void a_Nav_reload(BrowserWindow *bw)
 {
-   DilloUrl *url;
-   int choice;
-
-   a_Nav_cancel_expect(bw);
-   if (a_Nav_stack_size(bw)) {
-      url = a_History_get_url(NAV_TOP_UIDX(bw));
-      if (URL_FLAGS(url) & URL_Post) {
-         /* Attempt to repost data, let's confirm... */
-         choice = a_Dialog_choice3("Repost form data?",
-                                   "Yes", "*No", "Cancel");
-         if (choice == 0) { /* "Yes" */
-            _MSG("Nav_reload_confirmed\n");
-            Nav_reload(bw);
-         }
-
-      } else {
-         Nav_reload(bw);
-      }
-   }
+   dReturn_if_fail (bw != NULL);
+   a_Timeout_add(0.0, Nav_reload_callback, (void*)bw);
 }
 
 /*
@@ -536,10 +526,10 @@ void a_Nav_jump(BrowserWindow *bw, int offset, int new_bw)
    int idx = a_Nav_stack_ptr(bw) + offset;
 
    if (new_bw) {
-      a_Nav_push_nw(bw, a_History_get_url(NAV_UIDX(bw,idx)));
+      a_UIcmd_open_url_nw(bw, a_History_get_url(NAV_UIDX(bw,idx)));
    } else {
       a_Nav_cancel_expect(bw);
-      Nav_open_url(bw, a_History_get_url(NAV_UIDX(bw,idx)), offset);
+      Nav_open_url(bw, a_History_get_url(NAV_UIDX(bw,idx)), NULL, offset);
       a_UIcmd_set_buttons_sens(bw);
    }
 }
@@ -576,7 +566,7 @@ static void Nav_save_cb(int Op, CacheClient_t *Client)
 void a_Nav_save_url(BrowserWindow *bw,
                     const DilloUrl *url, const char *filename)
 {
-   DilloWeb *Web = a_Web_new(url);
+   DilloWeb *Web = a_Web_new(url, NULL);
    Web->bw = bw;
    Web->filename = dStrdup(filename);
    Web->flags |= WEB_Download;
@@ -599,3 +589,11 @@ void a_Nav_unref_buf(const DilloUrl *Url)
 {
    a_Capi_unref_buf(Url);
 }
+
+/*
+ * Wrapper for a_Capi_set_vsource_url().
+ */
+void a_Nav_set_vsource_url(const DilloUrl *Url)
+{
+   a_Capi_set_vsource_url(Url);
+}
