@@ -14,9 +14,8 @@ extern "C" {
 #endif /* __cplusplus */
 
 void a_Nav_redirection0(BrowserWindow *bw, const DilloUrl *new_url);
-void a_Nav_push(BrowserWindow *bw, const DilloUrl *url);
-void a_Nav_push_nw(BrowserWindow *bw, const DilloUrl *url);
-void a_Nav_vpush(void *vbw, const DilloUrl *url);
+void a_Nav_push(BrowserWindow *bw, const DilloUrl *url,
+                const DilloUrl *requester);
 void a_Nav_repush(BrowserWindow *bw);
 void a_Nav_back(BrowserWindow *bw);
 void a_Nav_forw(BrowserWindow *bw);
@@ -25,6 +24,7 @@ void a_Nav_reload(BrowserWindow *bw);
 void a_Nav_jump(BrowserWindow *bw, int offset, int new_bw);
 void a_Nav_free(BrowserWindow *bw);
 void a_Nav_cancel_expect (BrowserWindow *bw);
+void a_Nav_cancel_expect_if_eq(BrowserWindow *bw, const DilloUrl *url);
 void a_Nav_expect_done(BrowserWindow *bw);
 int a_Nav_stack_ptr(BrowserWindow *bw);
 int a_Nav_stack_size(BrowserWindow *bw);
@@ -35,6 +35,7 @@ void a_Nav_save_url(BrowserWindow *bw,
                     const DilloUrl *url, const char *filename);
 int a_Nav_get_buf(const DilloUrl *Url, char **PBuf, int *BufSize);
 void a_Nav_unref_buf(const DilloUrl *Url);
+void a_Nav_set_vsource_url(const DilloUrl *Url);
 
 #ifdef __cplusplus
 }
