@@ -11,6 +11,19 @@
 
 #include "prefs.h"
 
+#define PREFS_START_PAGE      "about:splash"
+#define PREFS_HOME            "http://www.dillo.org/"
+#define PREFS_FONT_SERIF      "DejaVu Serif"
+#define PREFS_FONT_SANS_SERIF "DejaVu Sans"
+#define PREFS_FONT_CURSIVE    "URW Chancery L"
+#define PREFS_FONT_FANTASY    "DejaVu Sans" /* TODO: find good default */
+#define PREFS_FONT_MONOSPACE  "DejaVu Sans Mono"
+#define PREFS_SEARCH_URL      "http://duckduckgo.com/lite/?kp=-1&q=%s"
+#define PREFS_NO_PROXY        "localhost 127.0.0.1"
+#define PREFS_SAVE_DIR        "/tmp/"
+#define PREFS_HTTP_REFERER    "host"
+#define PREFS_HTTP_USER_AGENT "Dillo/" VERSION
+
 /*-----------------------------------------------------------------------------
  * Global Data
  *---------------------------------------------------------------------------*/
@@ -23,9 +36,11 @@ DilloPrefs prefs;
 void a_Prefs_init(void)
 {
    prefs.allow_white_bg = TRUE;
+   prefs.bg_color = 0xdcd1ba;
    prefs.buffered_drawing = 1;
    prefs.contrast_visited_color = TRUE;
    prefs.enterpress_forces_submit = FALSE;
+   prefs.filter_auto_requests = PREFS_FILTER_SAME_DOMAIN;
    prefs.focus_new_tab = TRUE;
    prefs.font_cursive = dStrdup(PREFS_FONT_CURSIVE);
    prefs.font_factor = 1.0;
@@ -46,24 +61,31 @@ void a_Prefs_init(void)
    prefs.home = a_Url_new(PREFS_HOME, NULL);
    prefs.http_language = NULL;
    prefs.http_proxy = NULL;
+   prefs.http_max_conns = 6;
    prefs.http_proxyuser = NULL;
    prefs.http_referer = dStrdup(PREFS_HTTP_REFERER);
+   prefs.http_user_agent = dStrdup(PREFS_HTTP_USER_AGENT);
    prefs.limit_text_width = FALSE;
    prefs.load_images=TRUE;
    prefs.load_stylesheets=TRUE;
    prefs.middle_click_drags_page = TRUE;
    prefs.middle_click_opens_new_tab = TRUE;
-   prefs.no_proxy = NULL;
+   prefs.right_click_closes_tab = FALSE;
+   prefs.no_proxy = dStrdup(PREFS_NO_PROXY);
    prefs.panel_size = P_medium;
    prefs.parse_embedded_css=TRUE;
    prefs.save_dir = dStrdup(PREFS_SAVE_DIR);
-   prefs.search_url = dStrdup(PREFS_SEARCH_URL);
+   prefs.search_urls = dList_new(16);
+   dList_append(prefs.search_urls, dStrdup(PREFS_SEARCH_URL));
+   dList_append(prefs.search_urls, NULL); /* flags a default search URL */
+   prefs.search_url_idx = 0;
    prefs.show_back = TRUE;
    prefs.show_bookmarks = TRUE;
    prefs.show_clear_url = TRUE;
    prefs.show_extra_warnings = FALSE;
    prefs.show_filemenu=TRUE;
    prefs.show_forw = TRUE;
+   prefs.show_help = TRUE;
    prefs.show_home = TRUE;
    prefs.show_msg = TRUE;
    prefs.show_progress_box = TRUE;
@@ -85,6 +107,8 @@ void a_Prefs_init(void)
  */
 void a_Prefs_freeall(void)
 {
+   int i;
+
    dFree(prefs.font_cursive);
    dFree(prefs.font_fantasy);
    dFree(prefs.font_monospace);
@@ -95,8 +119,11 @@ void a_Prefs_freeall(void)
    a_Url_free(prefs.http_proxy);
    dFree(prefs.http_proxyuser);
    dFree(prefs.http_referer);
+   dFree(prefs.http_user_agent);
    dFree(prefs.no_proxy);
    dFree(prefs.save_dir);
-   dFree(prefs.search_url);
+   for (i = 0; i < dList_length(prefs.search_urls); ++i)
+      dFree(dList_nth_data(prefs.search_urls, i));
+   dList_free(prefs.search_urls);
    a_Url_free(prefs.start_page);
 }
