@@ -22,19 +22,12 @@ extern "C" {
 #define PREFS_GEOMETRY_DEFAULT_HEIGHT  580
 #define PREFS_GEOMETRY_DEFAULT_XPOS  -9999
 #define PREFS_GEOMETRY_DEFAULT_YPOS  -9999
-#define PREFS_START_PAGE      "about:splash"
-#define PREFS_HOME            "http://www.dillo.org/"
-#define PREFS_FONT_SERIF      "DejaVu Serif"
-#define PREFS_FONT_SANS_SERIF "DejaVu Sans"
-#define PREFS_FONT_CURSIVE    "DejaVu Sans" /* TODO: find good default */
-#define PREFS_FONT_FANTASY    "DejaVu Sans" /* TODO: find good default */
-#define PREFS_FONT_MONOSPACE  "DejaVu Sans Mono"
-#define PREFS_SEARCH_URL "http://www.google.com/search?ie=UTF-8&oe=UTF-8&q=%s"
-#define PREFS_SAVE_DIR        "/tmp/"
-#define PREFS_HTTP_REFERER    "host"
 
 /* Panel sizes */
-enum { P_tiny = 0, P_small, P_medium, P_large };
+enum { P_tiny = 0, P_small, P_medium };
+
+enum {PREFS_FILTER_ALLOW_ALL,
+      PREFS_FILTER_SAME_DOMAIN};
 
 typedef struct _DilloPrefs DilloPrefs;
 
@@ -44,13 +37,16 @@ struct _DilloPrefs {
    int xpos;
    int ypos;
    char *http_language;
+   int32_t http_max_conns;
    DilloUrl *http_proxy;
    char *http_proxyuser;
    char *http_referer;
+   char *http_user_agent;
    char *no_proxy;
    DilloUrl *start_page;
    DilloUrl *home;
    bool_t allow_white_bg;
+   int32_t bg_color;
    bool_t contrast_visited_color;
    bool_t show_tooltip;
    int panel_size;
@@ -73,11 +69,13 @@ struct _DilloPrefs {
    bool_t show_clear_url;
    bool_t show_url;
    bool_t show_search;
+   bool_t show_help;
    bool_t show_progress_box;
    bool_t fullwindow_start;
    bool_t load_images;
    bool_t load_stylesheets;
    bool_t parse_embedded_css;
+   int filter_auto_requests;
    int32_t buffered_drawing;
    char *font_serif;
    char *font_sans_serif;
@@ -86,7 +84,9 @@ struct _DilloPrefs {
    char *font_monospace;
    bool_t enterpress_forces_submit;
    bool_t middle_click_opens_new_tab;
-   char *search_url;
+   bool_t right_click_closes_tab;
+   bool_t search_url_idx;
+   Dlist *search_urls;
    char *save_dir;
    bool_t show_msg;
    bool_t show_extra_warnings;
