@@ -61,10 +61,13 @@ int a_Web_dispatch_by_type (const char *Type, DilloWeb *Web,
    if (Web->flags & WEB_RootUrl) {
       /* We have RootUrl! */
 
+      style::Color *bgColor = style::Color::create (layout, prefs.bg_color);
+      Web->bgColor = bgColor->getColor ();
+      layout->setBgColor (bgColor);
+
       /* Set a style for the widget */
       StyleEngine styleEngine (layout);
       styleEngine.startElement ("body");
-      Web->bgColor= styleEngine.backgroundStyle()->backgroundColor->getColor();
 
       dw = (Widget*) a_Mime_set_viewer(Type, Web, Call, Data);
       if (dw == NULL)
@@ -75,8 +78,11 @@ int a_Web_dispatch_by_type (const char *Type, DilloWeb *Web,
       /* This method frees the old dw if any */
       layout->setWidget(dw);
 
-      /* Clear the title bar for pages without a <TITLE> tag */
-      a_UIcmd_set_page_title(Web->bw, "");
+      /* Set the page title with the bare filename (e.g. for images),
+       * HTML pages with a <TITLE> tag will overwrite it later */
+      const char *p = strrchr(URL_STR(Web->url), '/');
+      a_UIcmd_set_page_title(Web->bw, p ? p+1 : "");
+
       a_UIcmd_set_location_text(Web->bw, URL_STR(Web->url));
       /* Reset both progress bars */
       a_UIcmd_set_page_prog(Web->bw, 0, 2);
@@ -89,12 +95,12 @@ int a_Web_dispatch_by_type (const char *Type, DilloWeb *Web,
 
    } else {
       /* A non-RootUrl. At this moment we only handle image-children */
-      if (!dStrncasecmp(Type, "image/", 6))
+      if (!dStrncasecmp(Type, "image/", 6)) {
          dw = (Widget*) a_Mime_set_viewer(Type, Web, Call, Data);
-   }
-
-   if (!dw) {
-      MSG_HTTP("unhandled MIME type: \"%s\"\n", Type);
+      } else {
+         MSG_HTTP("'%s' cannot be displayed as image; has media type '%s'\n",
+                  URL_STR(Web->url), Type);
+      }
    }
    return (dw ? 1 : -1);
 }
@@ -103,12 +109,13 @@ int a_Web_dispatch_by_type (const char *Type, DilloWeb *Web,
 /*
  * Allocate and set safe values for a DilloWeb structure
  */
-DilloWeb* a_Web_new(const DilloUrl *url)
+DilloWeb* a_Web_new(const DilloUrl *url, const DilloUrl *requester)
 {
    DilloWeb *web= dNew(DilloWeb, 1);
 
    _MSG(" a_Web_new: ValidWebs ==> %d\n", dList_length(ValidWebs));
    web->url = a_Url_dup(url);
+   web->requester = a_Url_dup(requester);
    web->bw = NULL;
    web->flags = 0;
    web->Image = NULL;
@@ -136,9 +143,11 @@ void a_Web_free(DilloWeb *web)
 {
    if (!web) return;
    a_Url_free(web->url);
+   a_Url_free(web->requester);
    a_Image_unref(web->Image);
    dFree(web->filename);
    dList_remove(ValidWebs, (void *)web);
+   _MSG("a_Web_free: ValidWebs=%d\n", dList_length(ValidWebs));
    dFree(web);
 }
 
